# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
# Author: Matteo Guardiani, Philipp Frank

import numpy as np
import scipy.linalg as slg
import scipy.sparse.linalg as ssl

from .field import Field
from .linearization import Linearization
from .logger import logger
from .minimization.sample_list import ResidualSampleList, SampleList
from .operator_spectrum import _DomRemover
from .operators.energy_operators import StandardHamiltonian
from .operators.sandwich_operator import SandwichOperator
from .sugar import makeField


class _Projector(ssl.LinearOperator):
    """Computes the projector of a Matrix or LinearOperator as a LinearOperator
    given the eigenvectors of the complementary space.

    Parameters
    ----------
    eigenvectors : ndarray
        The eigenvectors representing the directions to project out.

    Returns
    -------
    Projector : LinearOperator
        Operator representing the projection.
    """

    def __init__(self, eigenvectors):
        super().__init__(eigenvectors.dtype, 2 * (eigenvectors.shape[0],))
        self.eigenvectors = eigenvectors

    def _matvec(self, x):
        res = x.copy()
        for eigenvector in self.eigenvectors.T:
            res -= eigenvector * np.vdot(eigenvector, x)
        return res

    def _rmatvec(self, x):
        return self._matvec(x)


def _explicify(M):
    identity = np.identity(M.shape[0], dtype=np.float64)
    m = []
    for v in identity:
        m.append(M.matvec(v))
    return np.column_stack(m)


def _eigsh(
    metric,
    n_eigenvalues,
    tot_dofs,
    dtype,
    min_lh_eval=1e-4,
    n_batches=10,
    tol=0.0,
    verbose=True,
):
    metric = SandwichOperator.make(_DomRemover(metric.domain).adjoint, metric)
    metric_size = metric.domain.size
    M = ssl.LinearOperator(
        shape=2 * (metric_size,),
        matvec=lambda x: metric(makeField(metric.domain, x)).asnumpy(),
        dtype=dtype,
    )
    eigenvectors = None

    if n_eigenvalues > tot_dofs:
        raise ValueError(
            "Number of requested eigenvalues exceeds the number of relevant degrees of freedom!"
        )

    if tot_dofs == n_eigenvalues:
        # Compute exact eigensystem
        if verbose:
            logger.info(f"Computing all {tot_dofs} relevant metric eigenvalues.")
        eigenvalues = slg.eigh(
            _explicify(M),
            eigvals_only=True,
            subset_by_index=[metric_size - tot_dofs, metric_size - 1],
        )
        eigenvalues = np.flip(eigenvalues)
    else:
        # Set up batches
        batch_size = n_eigenvalues // n_batches
        remainder = n_eigenvalues % batch_size
        batches = [
            batch_size,
        ] * n_batches
        batches += (
            [
                remainder,
            ]
            if remainder > 0
            else []
        )
        eigenvalues, projected_metric = None, M

        for batch in batches:
            if verbose:
                logger.info(f"\nNumber of eigenvalues being computed: {batch}")
            # Get eigensystem for current batch
            eigvals, eigvecs = ssl.eigsh(
                projected_metric, k=batch, tol=tol, return_eigenvectors=True, which="LM"
            )
            i = np.argsort(eigvals)
            eigvals, eigvecs = np.flip(eigvals[i]), np.flip(eigvecs[:, i], axis=1)
            eigenvalues = (
                eigvals
                if eigenvalues is None
                else np.concatenate((eigenvalues, eigvals))
            )
            eigenvectors = (
                eigvecs if eigenvectors is None else np.hstack((eigenvectors, eigvecs))
            )

            if abs(1.0 - np.min(eigenvalues)) < min_lh_eval:
                break
            # Project out subspace of already computed eigenvalues
            projector = _Projector(eigenvectors)
            projected_metric = projector @ M @ projector.T
    return eigenvalues, eigenvectors


def estimate_evidence_lower_bound(
    hamiltonian,
    samples,
    n_eigenvalues,
    *,
    min_lh_eval=1e-3,
    n_batches=10,
    tol=0.0,
    verbose=True,
    dtype=np.float64,
):
    """Provides an estimate for the Evidence Lower Bound (ELBO).

    Statistical inference deals with the problem of hypothesis testing, given
    some data and models that can describe it. In general, it is hard to find a
    good metric to discern between different models. In Bayesian Inference, the
    Bayes factor can serve this purpose. To compute the Bayes factor it is
    necessary to calculate the evidence, given the specific model :math:`p(
    d|\\text{model})` at hand. Then, the ratio between the evidence of a model
    A and the one of a model B represents how much more likely it is that model
    A represents the data better than model B.

    The evidence for an approximated-inference problem can in principle be
    calculated. However, this is only practically feasible in a low-dimensional
    setting. What often can be computed is

    .. math ::
        \\log(p(d)) - D_\\text{KL} \\left[ Q(\\theta(\\xi)|d) || p(\\theta(\\xi) | d) \\right] = -\\langle H(\\theta(
        \\xi), d)\\rangle + \\frac1 2 \\left( N + \\text{Tr } \\log\\Lambda\\right),

    where :math:`D_\\text{KL} \\left[ Q || p \\right]` is the Kullback-Leibler
    (KL) divergence between the approximating posterior distribution :math:`Q`
    and the actual posterior :math:`p`. Since the Kullback-Leibler divergence
    :math:`D_\\text{KL} [\\cdot, \\cdot] \\geq 0` is positive definite, it is
    convenient to consider the lower bound

    .. math ::
        \\log(p(d)) \\geq -\\langle H(\\theta(\\xi), d)\\rangle + \\frac1 2 \\left(N + \\text{Tr } \\log\\Lambda
        \\right),

    which takes the name of Evidence Lower Bound (ELBO).

    If the KL divergence is well minimized (which should always be the case
    when a Variational Inference approach is followed), then it is possible to
    utilize the ELBO (as a proxy for the actual evidences) and calculate the
    Bayes factors for model comparison.


    Parameters
    ----------
    hamiltonian : :class:`nifty.cl.operators.energy_operators.StandardHamiltonian`
        Hamiltonian of the approximated probability distribution.
    samples : ResidualSampleList
        Collection of samples from the posterior distribution.
    n_eigenvalues : int
        Maximum number of eigenvalues to be considered for the estimation of
        the log-determinant of the metric. Note that if `n_eigenvalues` equals
        the total number of relevant degrees of freedom of the problem, all
        relevant eigenvalues are always computed irrespective of other stopping
        criteria.
    min_lh_eval : float
        Smallest eigenvalue of the likelihood to be considered. If the
        estimated eigenvalues become smaller then 1 + `min_lh_eval`, the
        eigenvalue estimation terminates and uses the smallest eigenvalue as a
        proxy for all remaining eigenvalues in the trace-log estimation.
        Default is 1e-3.
    n_batches : int
        Number of batches into which the eigenvalue estimation gets subdivided
        into. Only after completing one batch the early stopping criterion
        based on `min_lh_eval` is checked for.
    tol : Optional[float]
        Tolerance on the eigenvalue calculation. Zero indicates machine
        precision. Default is 0.
    verbose : Optional[bool]
        Print list of eigenvalues and summary of evidence calculation. Default
        is True.
    dtype : Optional[numpy.dtype]
        Data type of the eigenvalues and eigenvectors. Default is np.float64.

    Returns
    -------
    `elbo_samples` : SampleList
        List of elbo samples from the posterior distribution. The samples are
        returned to allow for more accurate elbo statistics.
    stats : dict
        Dictionary with a summary of the statistics of the estimated ELBO.
        The keys of this dictionary are:

        - `elbo_mean`: returns the mean value of the elbo estimate calculated
          over posterior samples
        - `elbo_up`: returns an upper bound to the elbo estimate (given by one
          posterior-sample standard deviation)
        - `elbo_lw`: returns a lower bound to the elbo estimate (one standard
          deviation plus a maximal error on the metric trace-log)
        - `lower_error`: maximal error on the metric trace-log term given by
          the number of relevant metric eigenvalues different from 1 neglected
          in the estimation of the trace-log times the log of the smallest
          calculated eigenvalue.

    Warning
    -------
    To perform Variational Inference there is no need to take into account
    quantities that are not explicitly dependent on the signal. Explicitly
    calculating these terms can be expensive, therefore they are usually
    neglected. Since in most cases they are also not required for model
    comparison, the provided estimate may not include terms which are constant
    in the signal. Only when comparing models with different noise statistics
    (or when the ELBO is needed to approximate the true evidence) these
    contributions have to be considered. For example, for a Gaussian
    distributed signal and a linear problem (Wiener Filter problem) the only
    term missing is :math:`-\\frac1 2 \\log \\det |2 \\pi N|`, where :math:`N`
    is the noise covariance matrix.

    See also
    --------
    For further details we refer to:

    * Analytic formulation: P. Frank et al., Geometric Variational Inference <https://arxiv.org/pdf/2105.10470.pdf> (Sec. 5.1)
    * Conceptualization: A. Kostić et al. (manuscript in preparation).
    """
    if not isinstance(samples, ResidualSampleList):
        raise TypeError("samples attribute should be of type ResidualSampleList.")
    if not isinstance(hamiltonian, StandardHamiltonian):
        raise TypeError("hamiltonian is not an instance of `ift.StandardHamiltonian`.")

    n_data_points = (
        hamiltonian.likelihood_energy.data_domain.size
        if not None
        else hamiltonian.domain.size
    )
    n_relevant_dofs = min(
        n_data_points, hamiltonian.domain.size
    )  # Number of metric eigenvalues that are not one

    metric = hamiltonian(Linearization.make_var(samples._m, want_metric=True)).metric
    metric_size = metric.domain.size

    eigenvalues, _ = _eigsh(
        metric,
        n_eigenvalues,
        n_relevant_dofs,
        dtype,
        min_lh_eval=min_lh_eval,
        n_batches=n_batches,
        tol=tol,
        verbose=verbose,
    )
    if verbose:
        # FIXME
        logger.info(
            f"\nComputed {eigenvalues.size} largest eigenvalues (out of {n_relevant_dofs} relevant degrees of freedom)."
            f"\nThe remaining {metric_size - n_relevant_dofs} metric eigenvalues (out of {metric_size}) are equal to "
            f"1.\n\n{eigenvalues}."
        )

    # Return a list of ELBO samples and a summary of the ELBO statistics
    log_eigenvalues = np.log(eigenvalues)
    tr_log_lat_cov = -0.5 * np.sum(log_eigenvalues)
    tr_log_lat_cov_lower = (
        0.5 * (n_relevant_dofs - log_eigenvalues.size) * np.min(log_eigenvalues)
    )
    tr_log_lat_cov_lower = Field.scalar(tr_log_lat_cov_lower)
    posterior_contribution = Field.scalar(tr_log_lat_cov + 0.5 * metric_size)
    elbo_samples = SampleList(
        list(samples.iterator(lambda x: posterior_contribution - hamiltonian(x)))
    )

    stats = {"lower_error": tr_log_lat_cov_lower}
    elbo_mean, elbo_var = elbo_samples.sample_stat()
    elbo_up = elbo_mean + elbo_var.sqrt()
    elbo_lw = elbo_mean - elbo_var.sqrt() - stats["lower_error"]
    stats["elbo_mean"], stats["elbo_up"], stats["elbo_lw"] = elbo_mean, elbo_up, elbo_lw
    if verbose:
        s = (
            f"\nELBO decomposition (in log units)"
            f"\nELBO mean : {elbo_mean.asnumpy():.4e} "
            f"(upper: {elbo_up.asnumpy():.4e}, "
            f"lower: {elbo_lw.asnumpy():.4e})"
        )
        logger.info(s)

    return elbo_samples, stats



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import logging


def logger_init(level=logging.INFO):
    import logging

    res = logging.getLogger("NIFTy8")
    res.setLevel(level)
    res.propagate = False
    ch = logging.StreamHandler()
    ch.setLevel(level)
    res.addHandler(ch)
    return res


logger = logger_init()



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from .correlated_field import ICRField
from .grid import Grid, MGrid
from .grid_impl import HEALPixGrid, HPLogRGrid, LogGrid, SimpleOpenGrid, BrokenLogGrid, HPBrokenLogRGrid



# Re-implemantation of C based healpix reference: https://gitlab.mpcdf.mpg.de/mtr/healpix_reference

# /* -----------------------------------------------------------------------------
#  *
#  *  Copyright (C) 1997-2019 Krzysztof M. Gorski, Eric Hivon, Martin Reinecke,
#  *                          Benjamin D. Wandelt, Anthony J. Banday,
#  *                          Matthias Bartelmann,
#  *                          Reza Ansari & Kenneth M. Ganga
#  *
#  *  Implementation of the Healpix bare bones C library
#  *
#  *  Licensed under a 3-clause BSD style license - see LICENSE
#  *
#  *  For more information on HEALPix and additional software packages, see
#  *  https://healpix.sourceforge.io/
#  *
#  *  If you are using this code in your own packages, please consider citing
#  *  the original paper in your publications:
#  *  K.M. Gorski et al., 2005, Ap.J., 622, p.759
#  *  (http://adsabs.harvard.edu/abs/2005ApJ...622..759G)
#  *
#  *----------------------------------------------------------------------------*/

from functools import partial
import jax
import jax.experimental
import jax.experimental.checkify
from jax.tree_util import register_pytree_node_class
import jax.numpy as jnp

from dataclasses import dataclass
import numpy as np


PI = 3.141592653589793238462643383279502884197
JRLL = jnp.array([2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4])
JPLL = jnp.array([1, 3, 5, 7, 0, 2, 4, 6, 1, 3, 5, 7])

NB_XOFFSET = np.array([-1, -1, 0, 1, 1, 1, 0, -1])
NB_YOFFSET = np.array([0, 1, 1, 1, 0, -1, -1, -1])
NB_FACEARRAY = jnp.array(
    [
        [8, 9, 10, 11, -1, -1, -1, -1, 10, 11, 8, 9],  # S
        [5, 6, 7, 4, 8, 9, 10, 11, 9, 10, 11, 8],  # SE
        [-1, -1, -1, -1, 5, 6, 7, 4, -1, -1, -1, -1],  # E
        [4, 5, 6, 7, 11, 8, 9, 10, 11, 8, 9, 10],  # SW
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],  # center
        [1, 2, 3, 0, 0, 1, 2, 3, 5, 6, 7, 4],  # NE
        [-1, -1, -1, -1, 7, 4, 5, 6, -1, -1, -1, -1],  # W
        [3, 0, 1, 2, 3, 0, 1, 2, 4, 5, 6, 7],  # NW
        [2, 3, 0, 1, -1, -1, -1, -1, 0, 1, 2, 3],  # N
    ]
)
NB_SWAPARRAY = jnp.array(
    [
        [0, 0, 3],  # S
        [0, 0, 6],  # SE
        [0, 0, 0],  # E
        [0, 0, 5],  # SW
        [0, 0, 0],  # center
        [5, 0, 0],  # NE
        [0, 0, 0],  # W
        [6, 0, 0],  # NW
        [3, 0, 0],  # N
    ]
)


BAD_NBR_SLCT = np.array([0, 0, 2, 4, 4, 6, 6, 0])
BAD_NBR_SHFT = np.array([0, -1, 0, 1, 0, -2, 0, -1])

# Conversions between continuous coordinate systems


@register_pytree_node_class
@dataclass()
class T_loc:
    z: float
    s: float
    phi: float

    def tree_flatten(self):
        children = (self.z, self.s, self.phi)
        aux_data = None
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(*children)


# A structure describing the continuous Healpix coordinate system.
# f takes values in [0;11], x and y lie in [0.0; 1.0].
@register_pytree_node_class
@dataclass()
class T_hpc:
    x: float
    y: float
    f: np.int32

    def tree_flatten(self):
        children = (self.x, self.y, self.f)
        aux_data = None
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(*children)


# A structure describing a location on the sphere. Theta is the co-latitude
# in radians (0 at the North Pole, increasing to pi at the South Pole. Phi is
# the azimuth in radians.
@register_pytree_node_class
@dataclass()
class T_ang:
    theta: float
    phi: float

    def tree_flatten(self):
        children = (self.theta, self.phi)
        aux_data = None
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(*children)


# A structure describing a 3-vector with coordinates x, y and z.
@register_pytree_node_class
@dataclass()
class T_vec:
    x: float
    y: float
    z: float

    def tree_flatten(self):
        children = (self.x, self.y, self.z)
        aux_data = None
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(*children)

    def norm(self):
        return jnp.sqrt(self.x**2 + self.y**2 + self.z**2)


# A structure describing the discrete Healpix coordinate system.
# f takes values in [0;11], x and y lie in [0; nside[.
@register_pytree_node_class
@dataclass()
class T_hpd:
    x: np.int64
    y: np.int64
    f: np.int32

    def tree_flatten(self):
        children = (self.x, self.y, self.f)
        aux_data = None
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(*children)


def loc2hpc(loc: T_loc) -> T_hpc:
    def loc2hpc_mid(loc, tt):  # Equatorial region
        temp1 = 0.5 + tt  # [0.5; 4.5)
        temp2 = loc.z * 0.75  # [-0.5; +0.5]
        jp = temp1 - temp2  # index of ascending edge line [0; 5)
        jm = temp1 + temp2  # index of descending edge line [0; 5)
        ifp = jp.astype(np.int64)  # int(jp)  # in {0, 4}
        ifm = jm.astype(np.int64)  # int(jm)
        f = jax.lax.select_n(2 * (ifp == ifm) + (ifp < ifm), ifm + 8, ifp, (ifp | 4))
        return T_hpc(jm - ifm, 1 + ifp - jp, f)

    def loc2hpc_caps(loc, tt):
        ntt = jnp.minimum(tt.astype(np.int64), 3)

        tp = tt - ntt  # [0;1)
        tmp = loc.s / jnp.sqrt((1.0 + jnp.abs(loc.z)) * (1.0 / 3.0))

        jp = tp * tmp  # increasing edge line index
        jm = (1.0 - tp) * tmp  # decreasing edge line index
        jp = jnp.minimum(jp, 1.0)  # for points too close to the boundary
        jm = jnp.minimum(jm, 1.0)
        cond = loc.z >= 0
        x = jax.lax.select(cond, 1.0 - jm, jp)
        y = jax.lax.select(cond, 1.0 - jp, jm)
        f = jax.lax.select(cond, ntt, ntt + 8)
        return T_hpc(x, y, f)

    tt = 4.0 * ((loc.phi * (1.0 / (2.0 * PI))) % 1)
    return jax.lax.cond(jnp.abs(loc.z) <= 2.0 / 3.0, loc2hpc_mid, loc2hpc_caps, loc, tt)


def hpc2loc(hpc: T_hpc) -> T_loc:
    def c0(jr, hpc):
        tmp = jr * jr * (1.0 / 3.0)
        phi = (PI * 0.25) * (JPLL[hpc.f] + (hpc.x - hpc.y) / jr)
        return T_loc(1.0 - tmp, jnp.sqrt(tmp * (2.0 - tmp)), phi)

    def c1(jr, hpc):
        jr = 4.0 - jr
        tmp = jr * jr * (1.0 / 3.0)
        phi = (PI * 0.25) * (JPLL[hpc.f] + (hpc.x - hpc.y) / jr)
        return T_loc(tmp - 1.0, jnp.sqrt(tmp * (2.0 - tmp)), phi)

    def c2(jr, hpc):
        z = (2.0 - jr) * (2.0 / 3.0)
        phi = (PI * 0.25) * (JPLL[hpc.f] + hpc.x - hpc.y)
        return T_loc(z, jnp.sqrt((1.0 + z) * (1.0 - z)), phi)

    def s2(jr, hpc):
        return jax.lax.cond(jr > 3.0, c1, c2, jr, hpc)

    jr = JRLL[hpc.f] - hpc.x - hpc.y
    return jax.lax.cond(jr < 1.0, c0, s2, jr, hpc)


def ang2loc(ang: T_ang) -> T_loc:
    cth, sth = jnp.cos(ang.theta), jnp.sin(ang.theta)
    cond = sth < 0
    phi = jax.lax.select(cond, ang.phi + PI, ang.phi)
    sth = jax.lax.select(cond, -sth, sth)
    return T_loc(z=cth, s=sth, phi=phi)


def loc2ang(loc: T_loc) -> T_ang:
    return T_ang(theta=jnp.arctan2(loc.s, loc.z), phi=loc.phi)


def vec2loc(vec: T_vec) -> T_loc:
    vlen = vec.norm()
    cth = vec.z / vlen
    sth = jnp.sqrt(vec.x**2 + vec.y**2) / vlen
    return T_loc(z=cth, s=sth, phi=jnp.arctan2(vec.y, vec.x))


def loc2vec(loc: T_loc) -> T_vec:
    return T_vec(x=loc.s * jnp.cos(loc.phi), y=loc.s * jnp.sin(loc.phi), z=loc.z)


def _ang2vec(ang: T_ang) -> T_vec:
    return loc2vec(ang2loc(ang))


def _vec2ang(vec: T_vec) -> T_ang:
    return T_ang(
        theta=jnp.atan2(jnp.sqrt(vec.x * vec.x + vec.y * vec.y), vec.z),
        phi=jnp.atan2(vec.y, vec.x),
    )


# Conversions between discrete coordinate systems


def isqrt(v: np.int64):
    res = jnp.sqrt(v + 0.5).astype(np.int64)

    def r2(res, v):
        res = jax.lax.select(res * res > v, res - 1, res)
        res = jax.lax.select(((res + 1) * (res + 1)) <= v, res + 1, res)
        return res

    return jax.lax.cond(v < (1 << 50), lambda x, v: x, r2, res, v)


def c_idiv(a, b):
    # mimic C like integer division int(a / b)
    res = a / b
    return int(res) if isinstance(res, float) else res.astype(np.int64)


def spread_bits(v: np.int64):
    res = v & 0xFFFFFFFF
    res = (res ^ (res << 16)) & 0x0000FFFF0000FFFF
    res = (res ^ (res << 8)) & 0x00FF00FF00FF00FF
    res = (res ^ (res << 4)) & 0x0F0F0F0F0F0F0F0F
    res = (res ^ (res << 2)) & 0x3333333333333333
    res = (res ^ (res << 1)) & 0x5555555555555555
    return res


def compress_bits(v: np.int64):
    res = v & 0x5555555555555555
    res = (res ^ (res >> 1)) & 0x3333333333333333
    res = (res ^ (res >> 2)) & 0x0F0F0F0F0F0F0F0F
    res = (res ^ (res >> 4)) & 0x00FF00FF00FF00FF
    res = (res ^ (res >> 8)) & 0x0000FFFF0000FFFF
    res = (res ^ (res >> 16)) & 0x00000000FFFFFFFF
    return res


def hpd2nest(nside: np.int64, hpd: T_hpd) -> np.int64:
    return (hpd.f * nside * nside) + spread_bits(hpd.x) + (spread_bits(hpd.y) << 1)


def nest2hpd(nside: np.int64, pix: np.int64) -> T_hpd:
    npface_ = nside * nside
    p2 = pix & (npface_ - 1)
    return T_hpd(x=compress_bits(p2), y=compress_bits(p2 >> 1), f=c_idiv(pix, npface_))


def hpd2ring(nside_: np.int64, hpd: T_hpd) -> np.int64:
    jr = (JRLL[hpd.f] * nside_) - hpd.x - hpd.y - 1

    def bound(x, b):
        return jax.lax.select_n(
            2 * (x > b) + (x < 1),
            x,
            x + b,
            x - b,
        )

    def c0(hpd, jr, nside):
        jp = c_idiv(JPLL[hpd.f] * jr + hpd.x - hpd.y + 1, 2)
        jp = bound(jp, 4 * nside)
        return 2 * jr * (jr - 1) + jp - 1

    def c1(hpd, jr, nside):
        nl4 = 4 * nside
        jri = nl4 - jr
        jp = c_idiv(JPLL[hpd.f] * jri + hpd.x - hpd.y + 1, 2)
        jp = bound(jp, nl4)
        return 12 * nside * nside - 2 * (jri + 1) * jri + jp - 1

    def c2(hpd, jr, nside):
        nl4 = 4 * nside
        jp = c_idiv(JPLL[hpd.f] * nside + hpd.x - hpd.y + 1 + ((jr - nside) & 1), 2)
        jp = bound(jp, nl4)
        return 2 * nside * (nside - 1) + (jr - nside) * nl4 + jp - 1

    def s2(hpd, jr, nside):
        return jax.lax.cond(jr > 3 * nside_, c1, c2, hpd, jr, nside)

    return jax.lax.cond(jr < nside_, c0, s2, hpd, jr, nside_)


def ring2hpd(nside: np.int64, pix: np.int64):
    ncap = 2 * nside * (nside - 1)
    npix = 12 * nside * nside

    def c0(pix, nside, ncap, npix):  # North Polar cap
        iring = (1 + isqrt(1 + 2 * pix)) >> 1  # counted from North pole
        iphi = (pix + 1) - 2 * iring * (iring - 1)
        face = c_idiv(iphi - 1, iring)
        irt = iring - (JRLL[face] * nside) + 1
        ipt = 2 * iphi - JPLL[face] * iring - 1
        ipt = jax.lax.select(ipt >= 2 * nside, ipt - 8 * nside, ipt)
        return T_hpd((ipt - irt) >> 1, (-(ipt + irt)) >> 1, face)

    def c1(pix, nside, ncap, npix):  # Equatorial region
        ip = pix - ncap
        iring = (c_idiv(ip, 4 * nside)) + nside  # counted from North pole
        iphi = (ip % (4 * nside)) + 1
        kshift = (iring + nside) & 1
        ire = iring - nside + 1
        irm = 2 * nside + 2 - ire
        ifm = c_idiv(iphi - c_idiv(ire, 2) + nside - 1, nside)
        ifp = c_idiv(iphi - c_idiv(irm, 2) + nside - 1, nside)
        face = jax.lax.select_n(2 * (ifp == ifm) + (ifp < ifm), ifm + 8, ifp, ifp | 4)
        irt = iring - (JRLL[face] * nside) + 1
        ipt = 2 * iphi - JPLL[face] * nside - kshift - 1
        ipt = jax.lax.select(ipt >= 2 * nside, ipt - 8 * nside, ipt)
        return T_hpd((ipt - irt) >> 1, (-(ipt + irt)) >> 1, face)

    def c2(pix, nside, ncap, npix):  # South Polar cap
        ip = npix - pix
        iring = (1 + isqrt(2 * ip - 1)) >> 1  # counted from South pole
        iphi = 4 * iring + 1 - (ip - 2 * iring * (iring - 1))
        face = 8 + c_idiv(iphi - 1, iring)
        irt = 4 * nside - iring - (JRLL[face] * nside) + 1
        ipt = 2 * iphi - JPLL[face] * iring - 1
        ipt = jax.lax.select(ipt >= 2 * nside, ipt - 8 * nside, ipt)
        return T_hpd((ipt - irt) >> 1, (-(ipt + irt)) >> 1, face)

    def s2(pix, nside, ncap, npix):
        return jax.lax.cond(pix < (npix - ncap), c1, c2, pix, nside, ncap, npix)

    return jax.lax.cond(pix < ncap, c0, s2, pix, nside, ncap, npix)


def nest2ring(nside: np.int64, ipnest: np.int64) -> np.int64:
    if (nside & (nside - 1)) != 0:
        return -1
    return hpd2ring(nside, nest2hpd(nside, ipnest))


def ring2nest(nside: np.int64, ipring: np.int64) -> np.int64:
    if (nside & (nside - 1)) != 0:
        return -1
    return hpd2nest(nside, ring2hpd(nside, ipring))


# Mixed conversions


def loc2hpd(nside_: np.int64, loc: T_loc) -> T_hpd:
    tmp = loc2hpc(loc)
    return T_hpd(
        (tmp.x * nside_).astype(jnp.int64),
        (tmp.y * nside_).astype(jnp.int64),
        tmp.f,
    )


def hpd2loc(nside_: np.int64, hpd: T_hpd) -> T_loc:
    xns = 1.0 / nside_
    tmp = T_hpc(x=(hpd.x + 0.5) * xns, y=(hpd.y + 0.5) * xns, f=hpd.f)
    return hpc2loc(tmp)


def npix2nside(npix: np.int64) -> np.int64:
    res = isqrt(npix // 12)
    if res * res * 12 != npix:
        return -1
    return res


def nside2npix(nside: np.int64):
    return 12 * nside * nside


def ang2ring(nside: np.int64, ang: T_ang) -> np.int64:
    return hpd2ring(nside, loc2hpd(nside, ang2loc(ang)))


def ang2nest(nside: np.int64, ang: T_ang) -> np.int64:
    return hpd2nest(nside, loc2hpd(nside, ang2loc(ang)))


def vec2ring(nside: np.int64, vec: T_vec) -> np.int64:
    return hpd2ring(nside, loc2hpd(nside, vec2loc(vec)))


def vec2nest(nside: np.int64, vec: T_vec) -> np.int64:
    return hpd2nest(nside, loc2hpd(nside, vec2loc(vec)))


def ring2ang(nside: np.int64, ipix: np.int64) -> T_ang:
    return loc2ang(hpd2loc(nside, ring2hpd(nside, ipix)))


def nest2ang(nside: np.int64, ipix: np.int64) -> T_ang:
    return loc2ang(hpd2loc(nside, nest2hpd(nside, ipix)))


def ring2vec(nside: np.int64, ipix: np.int64) -> T_vec:
    return loc2vec(hpd2loc(nside, ring2hpd(nside, ipix)))


def nest2vec(nside: np.int64, ipix: np.int64) -> T_vec:
    return loc2vec(hpd2loc(nside, nest2hpd(nside, ipix)))


def nside2order(nside: np.int64) -> np.int64:
    # Round as jnp.log may use inconsistent arithmetic
    return jax.lax.select(
        (nside & (nside - 1)) == 0, jnp.round(jnp.log2(nside)).astype(jnp.int64), -1
    )


def face_neighbors_nest(nside, hpd):
    fpix = hpd.f << (2 * nside2order(nside))
    px0 = spread_bits(hpd.x)
    py0 = spread_bits(hpd.y) << 1
    pxp = spread_bits(hpd.x + 1)
    pyp = spread_bits(hpd.y + 1) << 1
    pxm = spread_bits(hpd.x - 1)
    pym = spread_bits(hpd.y - 1) << 1

    result = jnp.array(
        [
            fpix + pxm + py0,
            fpix + pxm + pyp,
            fpix + px0 + pyp,
            fpix + pxp + pyp,
            fpix + pxp + py0,
            fpix + pxp + pym,
            fpix + px0 + pym,
            fpix + pxm + pym,
        ],
        dtype=np.int64,
    )
    return result


def face_neighbors_ring(nside, hpd):
    result = list(
        hpd2ring(
            nside, T_hpd(x=hpd.x + NB_XOFFSET[m], y=hpd.y + NB_YOFFSET[m], f=hpd.f)
        )
        for m in range(8)
    )
    return jnp.array(result)


def edge_neighbors(nside, hpd, fun):
    def bound(xx, num, shift):
        cond = 2 * (xx < 0) + (xx >= nside)
        xx = jax.lax.select_n(cond, xx, xx - nside, xx + nside)
        num = jax.lax.select_n(cond, num, num + shift, num - shift)
        return xx, num

    result = []
    for i in range(8):
        x = hpd.x + NB_XOFFSET[i]
        y = hpd.y + NB_YOFFSET[i]
        nbnum = 4
        x, nbnum = bound(x, nbnum, 1)
        y, nbnum = bound(y, nbnum, 3)

        f = NB_FACEARRAY[nbnum, hpd.f]

        def nb():
            bits = NB_SWAPARRAY[nbnum, hpd.f >> 2]
            xx = jax.lax.select(bits & 1, nside - x - 1, x)
            yy = jax.lax.select(bits & 2, nside - y - 1, y)
            cc = bits & 4
            tm = xx
            xx = jax.lax.select(cc, yy, xx)
            yy = jax.lax.select(cc, tm, yy)
            return fun(nside, T_hpd(xx, yy, f))

        result.append(jax.lax.select(f >= 0, nb(), -1))
    return jnp.array(result)


def neighbors(nside, pix, nest=False):
    hpd = nest2hpd(nside, pix) if nest else ring2hpd(nside, pix)

    face = face_neighbors_nest if nest else face_neighbors_ring
    edge = partial(edge_neighbors, fun=hpd2nest if nest else hpd2ring)

    nsm1 = nside - 1
    cond = (hpd.x > 0) * (hpd.x < nsm1) * (hpd.y > 0) * (hpd.y < nsm1)
    return jax.lax.cond(cond, face, edge, nside, hpd)


# Mimic healpy function signatures


def pix2ang(nside, ipix, nest=False, lonlat=False):
    if lonlat:
        msg = "Only co-latitude and longitude in radians supported for now!"
        raise NotImplementedError(msg)
    ang = nest2ang(nside, ipix) if nest else ring2ang(nside, ipix)
    return ang.theta, ang.phi


def pix2vec(nside, ipix, nest=False):
    vec = nest2vec(nside, ipix) if nest else ring2vec(nside, ipix)
    return vec.x, vec.y, vec.z


def ang2pix(nside, theta, phi, nest=False, lonlat=False):
    if lonlat:
        msg = "Only co-latitude and longitude in radians supported for now!"
        raise NotImplementedError(msg)
    ang = T_ang(theta=theta, phi=phi)
    return ang2nest(nside, ang) if nest else ang2ring(nside, ang)


def vec2pix(nside, x, y, z, nest=False):
    vec = T_vec(x=x, y=y, z=z)
    return vec2nest(nside, vec) if nest else vec2ring(nside, vec)


def vec2ang(x, y, z, lonlat=False):
    if lonlat:
        msg = "Only co-latitude and longitude in radians supported for now!"
        raise NotImplementedError(msg)
    vec = T_vec(x=x, y=y, z=z)
    ang = _vec2ang(vec)
    return ang.theta, ang.phi


def ang2vec(theta, phi, lonlat=False):
    if lonlat:
        msg = "Only co-latitude and longitude in radians supported for now!"
        raise NotImplementedError(msg)
    ang = T_ang(theta, phi)
    vec = _ang2vec(ang)
    return vec.x, vec.y, vec.z


def get_all_neighbours(nside, theta_or_pix, phi=None, nest=False, lonlat=False):
    if lonlat:
        msg = "Only co-latitude and longitude in radians supported for now!"
        raise NotImplementedError(msg)
    if phi is not None:
        theta_or_pix = ang2pix(nside, theta_or_pix, phi, nest=nest, lonlat=lonlat)
    return neighbors(nside, theta_or_pix, nest=nest)


def get_all_neighbours_valid(nside, theta_or_pix, phi=None, nest=False, lonlat=False):
    nbrs = get_all_neighbours(nside, theta_or_pix, phi=phi, nest=nest, lonlat=lonlat)
    if not nest:
        nbrs = jax.vmap(ring2nest, (None, 0))(nside, nbrs)
    tmp = nbrs[BAD_NBR_SLCT] + BAD_NBR_SHFT
    nbrs = jax.lax.select(nbrs != -1, nbrs, tmp)
    if not nest:
        nbrs = jax.vmap(nest2ring, (None, 0))(nside, nbrs)
    return nbrs



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

###############################################################################
# Meanfield and fullcovariance variational inference
#
# The signal is a 1-D lognormal distributed field.
# The  data follows a Poisson likelihood.
# The posterior distribution is approximated with a diagonal, as well as a
# full covariance Gaussian distribution. This is achieved by minimizing
# a stochastic estimate of the KL-Divergence
#
# Note that the fullcovariance approximation scales quadratically with the
# number of parameters.
###############################################################################

import numpy as np
from matplotlib import pyplot as plt

import nifty.cl as ift

ift.random.push_sseq_from_seed(27)


if __name__ == "__main__":
    # Space and model setup
    position_space = ift.RGSpace([100])
    harmonic_space = position_space.get_default_codomain()
    HT = ift.HarmonicTransformOperator(harmonic_space, position_space)
    p_space = ift.PowerSpace(harmonic_space)

    pd = ift.PowerDistributor(harmonic_space, p_space)
    a = ift.PS_field(p_space, lambda k: 1.0 / (1.0 + k ** 2))
    A = pd(a)
    sky = 10 * ift.exp(HT(ift.makeOp(A))).ducktape("xi")
    R = ift.GeometryRemover(position_space)

    mask = np.zeros(position_space.shape)
    mask[mask.shape[0]//3:2*mask.shape[0]//3] = 1
    mask = ift.Field.from_raw(position_space, mask)
    R = ift.MaskOperator(mask)

    d_space = R.target[0]
    lamb = R(sky)

    # Generate simulated signal and data and build likelihood energy
    mock_position = ift.from_random(sky.domain, "normal")
    data = ift.random.current_rng().poisson(lamb(mock_position).asnumpy())
    data = ift.makeField(d_space, data)
    likelihood_energy = ift.PoissonianEnergy(data) @ lamb
    H = ift.StandardHamiltonian(likelihood_energy)

    # Settings for minimization
    IC = ift.StochasticAbsDeltaEnergyController(5, iteration_limit=200,
                                                name='advi')
    minimizer_fc = ift.ADVIOptimizer(IC, eta=0.1)
    minimizer_mf = ift.ADVIOptimizer(IC)

    # Initial positions
    position_fc = ift.from_random(H.domain)*0.1
    position_mf = ift.from_random(H.domain)*0.1

    # Setup of the variational models
    fc = ift.FullCovarianceVI(position_fc, H, 3, True, initial_sig=0.01)
    mf = ift.MeanFieldVI(position_mf, H, 3, True, initial_sig=0.01)


    niter = 10
    for ii in range(niter):
        # Plotting
        plt.plot(sky(fc.mean).asnumpy(), "b-", label="Full covariance")
        plt.plot(sky(mf.mean).asnumpy(), "r-", label="Mean field")
        for _ in range(5):
            plt.plot(sky(fc.draw_sample()).asnumpy(), "b-", alpha=0.3)
            plt.plot(sky(mf.draw_sample()).asnumpy(), "r-", alpha=0.3)
        plt.plot(R.adjoint(data).asnumpy(), "kx")
        plt.plot(sky(mock_position).asnumpy(), "k-", label="Ground truth")
        plt.legend()
        plt.ylim(0.1, data.asnumpy().max() + 10)
        fname = f"meanfield_{ii:03d}.png"
        plt.savefig(fname)
        print(f"Saved results as '{fname}' ({ii}/{niter-1}).")
        plt.close()
        # /Plotting

        # Run minimization
        fc.minimize(minimizer_fc)
        mf.minimize(minimizer_mf)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from .structured_domain import StructuredDomain


class PowerSpace(StructuredDomain):
    """Represents non-equidistantly binned spaces for power spectra.

    A power space is the result of a projection of a harmonic domain where
    k-modes of equal length get mapped to one power index.

    Parameters
    ----------
    harmonic_partner : StructuredDomain
        The harmonic domain of which this is the power space.
    binbounds : None, or tuple of float
        By default (binbounds=None):
            There are as many bins as there are distinct k-vector lengths in
            the harmonic partner space.
            The `binbounds` property of the PowerSpace will be None.
        else:
            The bin bounds requested for this PowerSpace. The array
            must be sorted and strictly ascending. The first entry is the right
            boundary of the first bin, and the last entry is the left boundary
            of the last bin, i.e. there will be `len(binbounds)+1` bins in
            total, with the first and last bins reaching to -+infinity,
            respectively.
    """

    _powerIndexCache = {}
    _needed_for_hash = ["_harmonic_partner", "_binbounds"]

    @staticmethod
    def linear_binbounds(nbin, first_bound, last_bound):
        """Produces linearly spaced bin bounds.

        Parameters
        ----------
        nbin : int
            the number of bins
        first_bound, last_bound : float
            the k values for the right boundary of the first bin and the left
            boundary of the last bin, respectively. They are given in length
            units of the harmonic partner space.

        Returns
        -------
        numpy.ndarray(numpy.float64)
            binbounds array with nbin-1 entries with
            binbounds[0]=first_bound and binbounds[-1]=last_bound and the
            remaining values equidistantly spaced (in linear scale) between
            these two.
        """
        nbin = int(nbin)
        if nbin < 3:
            raise ValueError("nbin must be at least 3")
        return np.linspace(float(first_bound), float(last_bound), nbin-1)

    @staticmethod
    def logarithmic_binbounds(nbin, first_bound, last_bound):
        """Produces logarithmically spaced bin bounds.

        Parameters
        ----------
        nbin : int
            the number of bins
        first_bound, last_bound : float
            the k values for the right boundary of the first bin and the left
            boundary of the last bin, respectively. They are given in length
            units of the harmonic partner space.

        Returns
        -------
        numpy.ndarray(numpy.float64)
            binbounds array with nbin-1 entries with
            binbounds[0]=first_bound and binbounds[-1]=last_bound and the
            remaining values equidistantly spaced (in natural logarithmic
            scale) between these two.
        """
        nbin = int(nbin)
        if nbin < 3:
            raise ValueError("nbin must be at least 3")
        return np.logspace(np.log(float(first_bound)),
                           np.log(float(last_bound)),
                           nbin-1, base=np.e)

    @staticmethod
    def useful_binbounds(space, logarithmic, nbin=None):
        """Produces bin bounds suitable for a given domain.

        Parameters
        ----------
        space : StructuredDomain
            the domain for which the binbounds will be computed.
        logarithmic : bool
            If True bins will have equal size in linear space; otherwise they
            will have equal size in logarithmic space.
        nbin : int, optional
            the number of bins
            If None, the highest possible number of bins will be used

        Returns
        -------
        numpy.ndarray(numpy.float64)
            Binbounds array with `nbin-1` entries, if `nbin` is
            supplied, or the maximum number of entries that does not produce
            empty bins, if `nbin` is not supplied.
            The first and last bin boundary are inferred from `space`.
        """
        if not (isinstance(space, StructuredDomain) and space.harmonic):
            raise ValueError("first argument must be a harmonic space.")
        if logarithmic is None and nbin is None:
            return None
        nbin = None if nbin is None else int(nbin)
        logarithmic = bool(logarithmic)
        dists = space.get_unique_k_lengths()
        if len(dists) < 3:
            raise ValueError("Space does not have enough unique k lengths")
        lbound = 0.5*(dists[0]+dists[1])
        rbound = 0.5*(dists[-2]+dists[-1])
        dists[0] = lbound
        dists[-1] = rbound
        if logarithmic:
            dists = np.log(dists)
        binsz_min = np.max(np.diff(dists))
        nbin_max = int((dists[-1]-dists[0])/binsz_min)+2
        if nbin is None:
            nbin = nbin_max
        if nbin < 3:
            raise ValueError("nbin must be at least 3")
        if nbin > nbin_max:
            raise ValueError("nbin is too large")
        if logarithmic:
            return PowerSpace.logarithmic_binbounds(nbin, lbound, rbound)
        else:
            return PowerSpace.linear_binbounds(nbin, lbound, rbound)

    def __init__(self, harmonic_partner, binbounds=None):
        if not (isinstance(harmonic_partner, StructuredDomain) and
                harmonic_partner.harmonic):
            raise ValueError("harmonic_partner must be a harmonic space.")
        if harmonic_partner.scalar_dvol is None:
            raise ValueError("harmonic partner must have "
                             "scalar volume factors")
        self._harmonic_partner = harmonic_partner
        pdvol = harmonic_partner.scalar_dvol

        if binbounds is not None:
            binbounds = tuple(binbounds)
            if min(binbounds) < 0:
                raise ValueError('Negative binbounds encountered')

        key = (harmonic_partner, binbounds)
        if self._powerIndexCache.get(key) is None:
            k_length_array = self.harmonic_partner.get_k_length_array()
            if binbounds is None:
                tmp = harmonic_partner.get_unique_k_lengths()
                tbb = 0.5*(tmp[:-1]+tmp[1:])
            else:
                tbb = binbounds
            temp_pindex = np.searchsorted(tbb, k_length_array.asnumpy())
            nbin = len(tbb)+1
            temp_rho = np.bincount(temp_pindex.ravel(), minlength=nbin)
            if (temp_rho == 0).any():
                raise ValueError("empty bins detected")
            # The explicit conversion to float64 is necessary because bincount
            # sometimes returns its result as an integer array, even when
            # floating-point weights are present ...
            temp_k_lengths = np.bincount(temp_pindex.ravel(),
                weights=k_length_array.asnumpy().ravel(),
                minlength=nbin).astype(np.float64, copy=False)
            temp_k_lengths = temp_k_lengths / temp_rho
            temp_k_lengths.flags.writeable = False
            temp_pindex.flags.writeable = False
            temp_dvol = temp_rho*pdvol
            temp_dvol.flags.writeable = False
            self._powerIndexCache[key] = (binbounds, temp_pindex,
                                          temp_k_lengths, temp_dvol)

        (self._binbounds, self._pindex, self._k_lengths, self._dvol) = \
            self._powerIndexCache[key]

    def __repr__(self):
        return ("PowerSpace(harmonic_partner={}, binbounds={})"
                .format(self.harmonic_partner, self._binbounds))

    @property
    def harmonic(self):
        """bool : Always False for this class."""
        return False

    @property
    def shape(self):
        return self.k_lengths.shape

    @property
    def size(self):
        return self.shape[0]

    @property
    def scalar_dvol(self):
        return None

    @property
    def dvol(self):
        return self._dvol

    @property
    def harmonic_partner(self):
        """StructuredDomain : the harmonic domain associated with `self`."""
        return self._harmonic_partner

    @property
    def binbounds(self):
        """None or tuple of float : inner bin boundaries

        The boundaries between bins, starting with the right boundary of the
        first bin, up to the left boundary of the last bin.

        `None` is used to indicate natural binning.
        """
        return self._binbounds

    @property
    def pindex(self):
        """numpy.ndarray : bin indices

        Bin index for every pixel in the harmonic partner.
        """
        return self._pindex

    @property
    def k_lengths(self):
        """numpy.ndarray(float) : k-vector length for each bin."""
        return self._k_lengths

    def __reduce__(self):
        return (_unpicklePowerSpace, (self._harmonic_partner, self._binbounds))


def _unpicklePowerSpace(*args):
    return PowerSpace(*args)



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from functools import partial
from typing import Callable, Optional, Union

import jax
import numpy as np
from jax import numpy as jnp

from ..logger import logger
from ..misc import deprecated
from ..model import LazyModel
from ..multi_grid.kernel import refinement_matrices
from ..num import amend_unique_
from ..tree_math import ShapeWithDtype
from .chart import HEALPixChart
from .healpix_refine import cov_sqrt as cov_sqrt_hp
from .healpix_refine import refine as refine_hp
from .util import (
    RefinementMatrices,
    get_cov_from_loc,
    get_refinement_shapewithdtype,
)


def _matrices_naive(chart, kernel: Callable, depth: int):
    cov_from_loc = get_cov_from_loc(kernel)

    def cov_mat(lvl, idx_hp, idx_r):
        # `idx_r` is the left-most radial pixel of the to-be-refined slice
        # Extend `gc` and `gf` radially
        gc, gf = chart.get_coarse_fine_pair((idx_hp, idx_r), lvl)
        assert gf.shape[0] == chart.fine_size ** (chart.ndim + 1)

        coord = jnp.concatenate((gc, gf), axis=0)
        return cov_from_loc(coord, coord)

    def ref_mat_from_cov(cov):
        olf, ks = refinement_matrices(cov, chart.fine_size ** (chart.ndim + 1))
        if chart.ndim > 1:
            olf = olf.reshape(
                chart.fine_size**2,
                chart.fine_size,
                chart.coarse_size**2,
                chart.coarse_size,
            )
        return olf, ks

    def ref_mat(lvl, idx_hp, idx_r):
        return ref_mat_from_cov(cov_mat(lvl, idx_hp, idx_r))

    opt_lin_filter, kernel_sqrt = [], []
    for lvl in range(depth):
        pix_hp_idx = jnp.arange(chart.shape_at(lvl)[0])
        if chart.ndim == 1:
            pix_r_off = None
            vdist = jax.vmap(partial(ref_mat, lvl), in_axes=(0, 0))
        elif chart.ndim == 2:
            pix_r_off = jnp.arange(chart.shape_at(lvl)[1] - chart.coarse_size + 1)
            vdist = jax.vmap(partial(ref_mat, lvl), in_axes=(None, 0))
            vdist = jax.vmap(vdist, in_axes=(0, None))
        else:
            raise AssertionError()
        olf, ks = vdist(pix_hp_idx, pix_r_off)
        opt_lin_filter.append(olf)
        kernel_sqrt.append(ks)

    return RefinementMatrices(
        opt_lin_filter, kernel_sqrt, None, (None,) * len(opt_lin_filter)
    )


def _matrices_tol(
    chart,
    kernel: Callable,
    depth: int,
    *,
    atol: Optional[float] = None,
    rtol: Optional[float] = None,
    which: Optional[str] = "dist",
    mat_buffer_size: Optional[int] = None,
    _verbosity: int = 0,
):
    dist_mat_from_loc = get_cov_from_loc(lambda x: x, None)
    if which not in (None, "dist", "cov"):
        ve = f"expected `which` to be one of 'dist' or 'cov'; got {which!r}"
        raise ValueError(ve)
    if atol is not None and rtol is not None and mat_buffer_size is None:
        raise TypeError("must specify `mat_buffer_size`")

    def dist_mat(lvl, idx_hp, idx_r):
        # `idx_r` is the left-most radial pixel of the to-be-refined slice
        # Extend `gc` and `gf` radially
        gc, gf = chart.get_coarse_fine_pair((idx_hp, idx_r), lvl)
        assert gf.shape[0] == chart.fine_size ** (chart.ndim + 1)

        coord = jnp.concatenate((gc, gf), axis=0)
        return dist_mat_from_loc(coord, coord)

    def ref_mat(cov):
        olf, ks = refinement_matrices(cov, chart.fine_size ** (chart.ndim + 1))
        if chart.ndim > 1:
            olf = olf.reshape(
                chart.fine_size**2,
                chart.fine_size,
                chart.coarse_size**2,
                chart.coarse_size,
            )
        return olf, ks

    opt_lin_filter, kernel_sqrt, idx_map = [], [], []
    for lvl in range(depth):
        pix_hp_idx = jnp.arange(chart.shape_at(lvl)[0])
        assert chart.ndim == 2
        pix_r_off = jnp.arange(chart.shape_at(lvl)[1] - chart.coarse_size + 1)
        # Map only over the radial axis b/c that is irregular anyways simply
        # due to the HEALPix geometry and manually scan over the HEALPix axis
        # to only save unique values
        vdist = jax.vmap(partial(dist_mat, lvl), in_axes=(None, 0))

        # Successively amend the duplicate-free distance/covariance matrices
        d = jax.eval_shape(vdist, pix_hp_idx[0], pix_r_off)
        u = jnp.full((mat_buffer_size,) + d.shape, jnp.nan, dtype=d.dtype)

        def scanned_amend_unique(u, pix):
            d = vdist(pix, pix_r_off)
            d = kernel(d) if which == "cov" else d
            if _verbosity > 1:
                # Magic code to move up curser by one line and delete whole line
                msg = "\x1b[1A\x1b[2K{pix}/{n}"
                jax.debug.print(msg, pix=pix, n=pix_hp_idx[-1])
            return amend_unique_(u, d, axis=0, atol=atol, rtol=rtol)

        if _verbosity > 1:
            jax.debug.print("")
        u, inv = jax.lax.scan(scanned_amend_unique, u, pix_hp_idx)
        # Cut away the placeholder for preserving static shapes
        n = np.unique(inv).size
        if n >= u.shape[0] or not np.all(np.isnan(u[n:])):
            raise ValueError("`mat_buffer_size` too small")
        u = u[:n]
        u = kernel(u) if which == "dist" else u
        inv = np.array(inv)
        if _verbosity > 0:
            logger.info(f"Post uniquifying: {u.shape}")

        # Finally, all distance/covariance matrices are assembled and we
        # can map over them to construct the refinement matrices as usual
        vmat = jax.vmap(jax.vmap(ref_mat, in_axes=(0,)), in_axes=(0,))
        olf, ks = vmat(u)

        opt_lin_filter.append(olf)
        kernel_sqrt.append(ks)
        idx_map.append(inv)

    return RefinementMatrices(opt_lin_filter, kernel_sqrt, None, idx_map)


class ChartedHPField(LazyModel):
    @deprecated("`ChartedHPField` is deprecated in favor of `ICRField`")
    def __init__(
        self,
        chart: HEALPixChart,
        kernel: Optional[Callable] = None,
        dtype=None,
        skip0: bool = False,
    ):
        """Initialize an Iterative Charted Refinement (ICR) field for a HEALPix
        map with a radial extent.

        Parameters
        ----------
        chart :
            HEALPix coordinate chart with which to iteratively refine.
        kernel :
            Covariance kernel of the refinement field.
        dtype :
            Data-type of the excitations which to add during refining.
        skip0 :
            Whether to skip the first refinement level. This is useful to e.g.
        """
        self._chart = chart
        self._kernel = kernel
        self._dtype = dtype
        self._skip0 = skip0

    @property
    def kernel(self):
        """Yields the kernel specified during initialization or throw a
        `TypeError`.
        """
        if self._kernel is None:
            te = (
                "either specify a fixed kernel during initialization of the"
                f" {self.__class__.__name__} class or provide one here"
            )
            raise TypeError(te)
        return self._kernel

    @property
    def dtype(self):
        """Yields the data-type of the excitations."""
        return jnp.float64 if self._dtype is None else self._dtype

    @property
    def skip0(self):
        """Whether to skip the zeroth refinement"""
        return self._skip0

    @property
    def chart(self):
        """Associated `HEALPixChart` with which to iteratively refine."""
        return self._chart

    def matrices(
        self,
        kernel: Optional[Callable] = None,
        depth: Optional[int] = None,
        skip0: Optional[bool] = None,
        *,
        atol: Optional[float] = None,
        rtol: Optional[float] = None,
        which: Optional[str] = "dist",
        mat_buffer_size: Optional[int] = None,
        _verbosity: int = 0,
    ) -> RefinementMatrices:
        """Computes the refinement matrices namely the optimal linear filter
        and the square root of the information propagator (a.k.a. the square
        root of the fine covariance matrix for the excitations) for all
        refinement levels and all pixel indices in the coordinate chart.

        Parameters
        ----------
        kernel :
            Covariance kernel of the refinement field if not specified during
            initialization.
        depth :
            Maximum refinement depth if different to the one of the
            `HEALPixChart`.
        skip0 :
            Whether to skip the first refinement level.
        atol :
            Absolute tolerance of the element-wise difference between distance
            matrices up to which refinement matrices are merged.
        rtol :
            Relative tolerance of the element-wise difference between distance
            matrices up to which refinement matrices are merged.
        which : 'cov' or 'dist'
            Type of the matrix for which the unique instances shall be found.
            If the covariance matrices (`cov`) are "uniquified", then the
            procedure depends on the kernel used. If the distance matrices
            (`dist`) are "uniquified", then the search for redundancies becomes
            independent of the kernel. For a fixed charting, the latter can
            always be done ahead of time while it might be easier to set
            sensible `atol` and `rtol` parameters for the latter.

        Note
        ----
        Finding duplicates in the distance matrices is computationally
        expensive if there are only few of them. However, if the chart is kept
        fixed, then this one time investment might still be worth it though.
        """
        kernel = self.kernel if kernel is None else kernel
        depth = self.chart.depth if depth is None else depth
        skip0 = self.skip0 if skip0 is None else skip0

        if atol is None and rtol is None:
            rfm = _matrices_naive(self.chart, kernel, depth)
        else:
            rfm = _matrices_tol(
                self.chart,
                kernel,
                depth,
                atol=atol,
                rtol=rtol,
                which=which,
                mat_buffer_size=mat_buffer_size,
                _verbosity=_verbosity,
            )
        cov_sqrt0 = cov_sqrt_hp(self.chart, kernel) if not skip0 else None
        return rfm._replace(cov_sqrt0=cov_sqrt0)

    @property
    def domain(self):
        """Yields the `ShapeWithDtype` of the primals."""
        nonhp_domain = get_refinement_shapewithdtype(
            shape0=self.chart.shape0[1:],
            depth=self.chart.depth,
            dtype=self.dtype,
            skip0=self.skip0,
            _coarse_size=self.chart.coarse_size,
            _fine_size=self.chart.fine_size,
            _fine_strategy=self.chart.fine_strategy,
        )
        if not self.skip0:
            domain = [
                ShapeWithDtype(
                    (12 * self.chart.nside0**2,) + nonhp_domain[0].shape,
                    nonhp_domain[0].dtype,
                )
            ]
        else:
            domain = [None]
        domain += [
            ShapeWithDtype(
                (12 * self.chart.nside_at(lvl) ** 2,)
                + swd.shape[:-1]
                + (4 * swd.shape[-1],),
                swd.dtype,
            )
            for lvl, swd in zip(range(self.chart.depth + 1), nonhp_domain[1:])
        ]
        return domain

    @staticmethod
    def apply(
        xi,
        chart: HEALPixChart,
        kernel: Union[Callable, RefinementMatrices],
        *,
        skip0: bool = False,
        _refine: Optional[Callable] = None,
        precision=None,
    ):
        """Static method to apply a refinement field given some excitations, a
        chart and a kernel.

        Parameters
        ----------
        xi :
            Latent parameters which to use for refining.
        chart :
            Chart with which to refine the non-HEALPix axis.
        kernel :
            Covariance kernel with which to build the refinement matrices.
        skip0 :
            Whether to skip the first refinement level.
        precision :
            See JAX's precision.
        """
        if xi[0].shape != chart.shape0:
            ve = "zeroth excitations do not fit to chart"
            raise ValueError(ve)

        if isinstance(kernel, RefinementMatrices):
            refinement = kernel
        else:
            refinement = ChartedHPField(chart, None, xi[0].dtype).matrices(
                kernel=kernel, skip0=skip0
            )
        refine_w_chart = partial(
            refine_hp if _refine is None else _refine, chart=chart, precision=precision
        )

        if not skip0:
            fine = (refinement.cov_sqrt0 @ xi[0].ravel()).reshape(xi[0].shape)
        else:
            if refinement.cov_sqrt0 is not None:
                raise AssertionError()
            fine = xi[0]
        for x, olf, k, im in zip(
            xi[1:], refinement.filter, refinement.propagator_sqrt, refinement.index_map
        ):
            fine = refine_w_chart(fine, x, olf, k, im)
        return fine

    def __call__(self, xi, kernel=None, *, skip0=None, **kwargs):
        """See `ChartedField.apply`."""
        kernel = self.kernel if kernel is None else kernel
        skip0 = self.skip0 if skip0 is None else skip0
        return self.apply(xi, self.chart, kernel=kernel, skip0=skip0, **kwargs)

    def __repr__(self):
        descr = f"{self.__class__.__name__}(chart={self.chart!r}"
        descr += f", kernel={self._kernel!r}" if self._kernel is not None else ""
        descr += f", dtype={self._dtype!r}" if self._dtype is not None else ""
        descr += f", skip0={self.skip0!r}" if self.skip0 is not False else ""
        descr += ")"
        return descr

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __hash__(self):
        return hash(repr(self))
